__name__ = "wellcomeml"
__version__ = "2020.9.0"
__description__ = """Utilities for managing nlp models and for
                     processing text-related data at Wellcome Data Labs"""
__url__ = "https://github.com/wellcometrust/wellcomeml/tree/master"
__author__ = "Wellcome Trust DataLabs Team"
__author_email__ = "Grp_datalabs-datascience@Wellcomecloud.onmicrosoft.com"
__license__ = "MIT"
