from .__version__ import (
    __name__, __version__, __description__, __url__,
    __author__, __author_email__, __license__
)
from .logger import logger

__all__ = [
    __name__, __version__, __description__, __url__,
    __author__, __author_email__, __license__, logger
]
